/******************************************************************************
  Copyright (c) 1995, 1996 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/*****************************************************************************
 * Routines for dumping and shutting down LambdaMOO databases
 *****************************************************************************/

#include "my-ctype.h"
#include <float.h>
#include "my-stat.h"
#include "my-stdarg.h"
#include "my-stdio.h"
#include "my-stdlib.h"

#include "config.h"
#include "db.h"
#include "decompile.h"
#include "exceptions.h"
#include "eval_env.h"
#include "eval_vm.h"
#include "functions.h"
#include "list.h"
#include "log.h"
#include "numbers.h"
#include "options.h"
#include "parser.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "str_intern.h"
#include "tasks.h"
#include "timers.h"
#include "unparse.h"
#include "utils.h"
#include "version.h"

#include "db_save_lambda_common.h"
 
static char *dump_db_name;
static int dump_generation = 0;

/*********** dbio ***********/


Exception dbpriv_dbio_failed;

static FILE *output;

void
dbpriv_set_dbio_output(FILE * f)
{
    output = f;
}

void
dbio_printf(const char *format,...)
{
    va_list args;

    va_start(args, format);
    if (vfprintf(output, format, args) < 0)
	RAISE(dbpriv_dbio_failed, 0);
    va_end(args);
}

void
dbio_write_num(int n)
{
    dbio_printf("%d\n", n);
}

void
dbio_write_float(double d)
{
    static const char *fmt = 0;
    static char buffer[10];

    if (!fmt) {
	sprintf(buffer, "%%.%dg\n", DBL_DIG + 4);
	fmt = buffer;
    }
    dbio_printf(fmt, d);
}

void
dbio_write_objid(Objid oid)
{
    dbio_write_num(oid);
}

void
dbio_write_string(const char *s)
{
	int pos;
	
	/* Handle null strings */
	if (!s) {
		fputc('\n', output);
		return;
	}

	/* Iterate over each character, printing to the file and Escaping as nessecary. */
	for (pos = 0; s[pos]; pos++)
	{
		switch (s[pos])
		{
			case '\n':
				fputc('\0', output);
				fputc('n', output);
				break;
			case '\r':
				fputc('\0', output);
				fputc('r', output);
				break;
			default:
				fputc(s[pos], output);
		}
	}
	fputc('\n', output);
	
    /* dbio_printf("%s\n", s ? s : ""); */
}

void
dbio_write_var(Var v)
{
    int i;

    dbio_write_num((int) v.type & TYPE_DB_MASK);
    switch ((int) v.type) {
    case TYPE_CLEAR:
    case TYPE_NONE:
	break;
    case TYPE_STR:
	dbio_write_string(v.v.str);
	break;
    case TYPE_OBJ:
    case TYPE_ERR:
    case TYPE_INT:
    case TYPE_CATCH:
    case TYPE_FINALLY:
	dbio_write_num(v.v.num);
	break;
    case TYPE_FLOAT:
	dbio_write_float(*v.v.fnum);
	break;
    case TYPE_LIST:
	dbio_write_num(v.v.list[0].v.num);
	for (i = 0; i < v.v.list[0].v.num; i++)
	    dbio_write_var(v.v.list[i + 1]);
	break;
    }
}

static void
receiver(void *data, const char *line)
{
    dbio_printf("%s\n", line);
}

void
dbio_write_program(Program * program)
{
    unparse_program(program, receiver, 0, 1, 0, MAIN_VECTOR);
    dbio_printf(".\n");
}

void
dbio_write_forked_program(Program * program, int f_index)
{
    unparse_program(program, receiver, 0, 1, 0, f_index);
    dbio_printf(".\n");
}

void
dbio_write_bidata(Var bidata)
{	// {fname, {vname, vval}, {vname, vval}, {"bidata", {f2name, {v2name, v2val},...}},...}
	// Has some checking for invalid data, but some may still go through (and possibly cause a crash),
	// so make sure all bidata is valid
	Stream *out;
	int i;
	char *ret;

	if (bidata.type == TYPE_INT)
		return;
	
	out = new_stream(40);
	stream_printf(out, "bf_%s data: ", bidata.v.list[1].v.str);
	for (i = 2; i <= bidata.v.list[0].v.num; ++i) {
		switch (bidata.v.list[i].type) {
		case TYPE_LIST:
			if (mystrcasecmp(bidata.v.list[i].v.list[1].v.str, "bidata")) {
				stream_add_string(out, bidata.v.list[i].v.list[1].v.str);
				switch (bidata.v.list[i].v.list[2].type) {
				case TYPE_INT:	break;
				case TYPE_OBJ:
					stream_printf(out, " = %d", bidata.v.list[i].v.list[2].v.obj);
					break;
				default:
					errlog("WARNING: Not writing unrecognized bidata value!\n");
				}
			}
			break;
		case TYPE_STR:
			stream_add_string(out, bidata.v.list[i].v.str);
			break;
		default:
			errlog("WARNING: Not writing unrecognized bidata item!\n");
			free_stream(out);
			return;
		}
		stream_add_string(out, ", ");
	}
	stream_del_multiple(out, 2);
	if (!mystrcasecmp(bidata.v.list[1].v.str, "recycle"))
		stream_add_string(out, ", cont = 0");	// Compatibility
	ret = str_dup(stream_contents(out));
	free_stream(out);
	dbio_printf("%s\n", ret);
	for (i = 2; i <= bidata.v.list[0].v.num; ++i)
		if (bidata.v.list[i].type == TYPE_LIST
		 && !mystrcasecmp(bidata.v.list[i].v.list[1].v.str, "bidata"))
			dbio_write_bidata(bidata.v.list[i].v.list[2]);
}

/*********** tasks ***********/

static void
write_forked_task(forked_task ft)
{
    int lineno = find_line_number(ft.program, ft.f_index, 0);

	/* saving rounds to the nearest second. restart is slow anyway */
    dbio_printf("0 %d %d %d\n", lineno, ROUND(&ft.start_tv), ft.id);
    write_activ_as_pi(ft.a);
    write_rt_env(ft.program->var_names, ft.rt_env, ft.program->num_var_names);
    dbio_write_forked_program(ft.program, ft.f_index);
}

static void
write_suspended_task(suspended_task st)
{
	/* saving rounds to the nearest second. restart is slow anyway */
    dbio_printf("%d %d ", ROUND(&st.start_tv), st.the_vm->task_id);
    dbio_write_var(st.value);
    write_vm(st.the_vm);
}

void
write_task_queue(void)
{
    int forked_count = 0;
    int suspended_count = 0;
    task *t;
    tqueue *tq;

    dbio_printf("0 clocks\n");	/* for compatibility's sake */

    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_FORKED)
	    forked_count++;
	else			/* t->kind == TASK_SUSPENDED */
	    suspended_count++;

    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first_bg; t; t = t->next)
	    if (t->kind == TASK_FORKED)
		forked_count++;
	    else		/* t->kind == TASK_SUSPENDED */
		suspended_count++;


    dbio_printf("%d queued tasks\n", forked_count);

    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_FORKED)
	    write_forked_task(t->t.forked);

    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first_bg; t; t = t->next)
	    if (t->kind == TASK_FORKED)
		write_forked_task(t->t.forked);

    dbio_printf("%d suspended tasks\n", suspended_count);

    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_SUSPENDED)
	    write_suspended_task(t->t.suspended);

    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first_bg; t; t = t->next)
	    if (t->kind == TASK_SUSPENDED)
		write_suspended_task(t->t.suspended);
}

/*********** server ***********/

void
write_active_connections(void)
{
    int count = 0;
    shandle *h;

    for (h = all_shandles; h; h = h->next)
	count++;

    dbio_printf("%d active connections with listeners\n", count);

    for (h = all_shandles; h; h = h->next)
	dbio_printf("%d %d\n", h->player, h->listener);
}

/*********** execute ***********/

void
write_activ_as_pi(activation a)
{
    Var dummy;

    dummy.type = TYPE_INT;
    dummy.v.num = -111;
    dbio_write_var(dummy);

    dbio_printf("%d %d %d %d %d %d %d %d %d\n",
	    a.this, -7, -8, a.player, -9, a.progr, a.vloc, -10, a.debug);
    dbio_write_string("No");
    dbio_write_string("More");
    dbio_write_string("Parse");
    dbio_write_string("Infos");
    dbio_write_string(a.verb);
    dbio_write_string(a.verbname);
}

void
write_rt_env(const char **var_names, Var * rt_env, unsigned size)
{
    unsigned i;

    dbio_printf("%d variables\n", size);
    for (i = 0; i < size; i++) {
	dbio_write_string(var_names[i]);
	dbio_write_var(rt_env[i]);
    }
}

void
write_activ(activation a)
{
    register Var *v;

    dbio_printf("language version %u\n", a.prog->version);
    dbio_write_program(a.prog);
    write_rt_env(a.prog->var_names, a.rt_env, a.prog->num_var_names);

    dbio_printf("%d rt_stack slots in use\n",
		a.top_rt_stack - a.base_rt_stack);

    for (v = a.base_rt_stack; v != a.top_rt_stack; v++)
	dbio_write_var(*v);

    write_activ_as_pi(a);
    dbio_write_var(a.temp);

    dbio_printf("%u %u %u\n", a.pc, a.bi_func_pc, a.error_pc);
	if (a.bi_func_pc != 0) {
		dbio_write_string(name_func_by_num(a.bi_func_id));
		dbio_write_bidata(get_bi_func_data(a.bi_func_data, a.bi_func_id));
    }
}

/*********** eval_vm ***********/

void
write_vm(vm the_vm)
{
    unsigned i;

    dbio_printf("%u %d %u %u\n",
		the_vm->top_activ_stack, the_vm->root_activ_vector,
		the_vm->func_id, the_vm->max_stack_size);

    for (i = 0; i <= the_vm->top_activ_stack; i++)
	write_activ(the_vm->activ_stack[i]);
}

/*********** Verb and property I/O ***********/

static void
write_verbdef(Verbdef * v)
{
    dbio_write_string(v->name);
    dbio_write_objid(v->owner);
    dbio_write_num(v->perms);
    dbio_write_num(v->prep);
}

static void
write_propdef(Propdef * p)
{
    dbio_write_string(p->name);
}

static void
write_propval(Pval * p)
{
    dbio_write_var(p->var);
    dbio_write_objid(p->owner);
    dbio_write_num(p->perms);
}


/*********** Object I/O ***********/

static void
write_object(Objid oid)
{
    Object *o;
    Verbdef *v;
    int i;
    int nverbdefs, nprops;

    if (!valid(oid)) {
	dbio_printf("#%d recycled\n", oid);
	return;
    }
    o = dbpriv_find_object(oid);

    dbio_printf("#%d\n", oid);
    dbio_write_string(o->name);
    dbio_write_string("");	/* placeholder for old handles string */
    dbio_write_num(o->flags);

    dbio_write_objid(o->owner);

    dbio_write_objid(o->location);
    dbio_write_objid(o->contents);
    dbio_write_objid(o->next);

    dbio_write_objid(o->parent);
    dbio_write_objid(o->child);
    dbio_write_objid(o->sibling);

    for (v = o->verbdefs, nverbdefs = 0; v; v = v->next)
	nverbdefs++;

    dbio_write_num(nverbdefs);
    for (v = o->verbdefs; v; v = v->next)
	write_verbdef(v);

    dbio_write_num(o->propdefs.cur_length);
    for (i = 0; i < o->propdefs.cur_length; i++)
	write_propdef(&o->propdefs.l[i]);

    nprops = dbpriv_count_properties(oid);

    dbio_write_num(nprops);
    for (i = 0; i < nprops; i++)
	write_propval(o->propval + i);
}


/*********** File-level Output ***********/

static int
write_db_file(const char *reason)
{
    Objid oid;
    Objid max_oid = db_last_used_objid();
    Verbdef *v;
    Var user_list;
    int i;
    volatile int nprogs = 0;
    volatile int success = 1;

    for (oid = 0; oid <= max_oid; oid++) {
	if (valid(oid))
	    for (v = dbpriv_find_object(oid)->verbdefs; v; v = v->next)
		if (v->program)
		    nprogs++;
    }

    user_list = db_all_users();

    TRY
	dbio_printf(header_format_string, current_version);
    dbio_printf("%d\n%d\n%d\n%d\n",
		max_oid + 1, nprogs, 0, user_list.v.list[0].v.num);
    for (i = 1; i <= user_list.v.list[0].v.num; i++)
	dbio_write_objid(user_list.v.list[i].v.obj);
    oklog("%s: Writing %d objects...\n", reason, max_oid + 1);
    for (oid = 0; oid <= max_oid; oid++) {
	write_object(oid);
	if ((oid + 1) % 10000 == 0 || oid == max_oid)
	    oklog("%s: Done writing %d objects...\n", reason, oid + 1);
    }
    oklog("%s: Writing %d MOO verb programs...\n", reason, nprogs);
    for (i = 0, oid = 0; oid <= max_oid; oid++)
	if (valid(oid)) {
	    int vcount = 0;

	    for (v = dbpriv_find_object(oid)->verbdefs; v; v = v->next) {
		if (v->program) {
		    dbio_printf("#%d:%d\n", oid, vcount);
		    dbio_write_program(v->program);
		    if (++i % 5000 == 0 || i == nprogs)
			oklog("%s: Done writing %d verb programs...\n",
			      reason, i);
		}
		vcount++;
	    }
	}
    oklog("%s: Writing forked and suspended tasks...\n", reason);
    write_task_queue();
    oklog("%s: Writing list of formerly active connections...\n", reason);
    write_active_connections();
    EXCEPT(dbpriv_dbio_failed)
	success = 0;
    ENDTRY;

    return success;
}

typedef enum {
    DUMP_SHUTDOWN, DUMP_CHECKPOINT, DUMP_PANIC
} Dump_Reason;
const char *reason_names[] =
{"DUMPING", "CHECKPOINTING", "PANIC-DUMPING"};

static int
dump_database_to(Dump_Reason reason, const char *filename)
{
    Stream *s = new_stream(100);
    char *temp_name;
    FILE *f;
    int success;

  retryDumping:

    stream_printf(s, "%s.#%d#", filename, dump_generation);
    remove(reset_stream(s));	/* Remove previous checkpoint */

    if (reason == DUMP_PANIC)
	stream_printf(s, "%s.PANIC", filename);
    else {
	dump_generation++;
	stream_printf(s, "%s.#%d#", filename, dump_generation);
    }
    temp_name = reset_stream(s);

    oklog("%s on %s ...\n", reason_names[reason], temp_name);

#ifdef UNFORKED_CHECKPOINTS
    reset_command_history();
#else
    if (reason == DUMP_CHECKPOINT) {
	switch (fork_server("checkpointer")) {
	case FORK_PARENT:
	    reset_command_history();
	    free_stream(s);
	    return 1;
	case FORK_ERROR:
	    free_stream(s);
	    return 0;
	case FORK_CHILD:
	    set_server_cmdline("(MOO checkpointer)");
	    break;
	}
    }
#endif

    success = 1;
    if ((f = fopen(temp_name, "w")) != 0) {
	dbpriv_set_dbio_output(f);
	if (!write_db_file(reason_names[reason])) {
	    log_perror("Trying to dump database");
	    fclose(f);
	    remove(temp_name);
	    if (reason == DUMP_CHECKPOINT) {
		errlog("Abandoning checkpoint attempt...\n");
		success = 0;
	    } else {
		int retry_interval = 60;

		errlog("Waiting %d seconds and retrying dump...\n",
		       retry_interval);
		timer_sleep(retry_interval);
		goto retryDumping;
	    }
	} else {
	    fclose(f);
	    oklog("%s on %s finished\n", reason_names[reason], temp_name);
	    if (reason != DUMP_PANIC) {
		remove(filename);
		if (rename(temp_name, filename) != 0) {
		    log_perror("Renaming temporary dump file");
		    success = 0;
		}
	    }
	}
    } else {
	log_perror("Opening temporary dump file");
	success = 0;
    }

    free_stream(s);

#ifndef UNFORKED_CHECKPOINTS
    if (reason == DUMP_CHECKPOINT)
	/* We're a child, so we'd better go away. */
	exit(!success);
#endif

    return success;
}

static int
dump_database(Dump_Reason reason)
{
	return dump_database_to(reason, dump_db_name);
}

/*********** External interface ***********/

const char *
dbflush_usage_string(void)
{
    return "output-db-file";
}

int
dbflush_initialize(int *pargc, char ***pargv)
{
    if (*pargc < 1)
	return 0;

    dump_db_name = str_dup((*pargv)[0]);
    --*pargc;
    ++*pargv;

    return 1;
}

int
db_flush(enum db_flush_type type)
{
    int success = 0;

    switch (type) {
    case FLUSH_IF_FULL:
    case FLUSH_ONE_SECOND:
	success = 1;
	break;

    case FLUSH_ALL_NOW:
	success = dump_database(DUMP_CHECKPOINT);
	break;

    case FLUSH_PANIC:
	success = dump_database(DUMP_PANIC);
	break;
    }

    return success;
}

int
db_flush_to(enum db_flush_type type, const char *filename)
{
    int success = 0;

    switch (type) {
    case FLUSH_IF_FULL:
    case FLUSH_ONE_SECOND:
	success = 1;
	break;

    case FLUSH_ALL_NOW:
	success = dump_database_to(DUMP_CHECKPOINT, filename);
	break;

    case FLUSH_PANIC:
	success = dump_database_to(DUMP_PANIC, filename);
	break;
    }

    return success;
}

void
dbflush_status(void)
{
    oklog("STATUS: Will dump new database on %s\n", dump_db_name);
}

int32
dbflush_disk_size(void)
{
    struct stat st;

	if (dump_generation == 0 || stat(dump_db_name, &st) < 0)
		return -1;
	else
		return st.st_size;
}

void
db_shutdown()
{
    dump_database(DUMP_SHUTDOWN);
}

char rcsid_db_save_lambda_flush[] = "$Id: db_save_lambda_flush.c,v 1.5 2004/10/10 16:37:36 luke-jr Exp $";

/* 
 * $Log: db_save_lambda_flush.c,v $
 * Revision 1.5  2004/10/10 16:37:36  luke-jr
 * more warn cleanups
 *
 * Revision 1.4  2004/03/20 12:17:03  luke-jr
 * db_options_v2 patch compatibility
 *
 * Revision 1.3  2004/03/18 23:46:37  luke-jr
 * main() called dbload/flush_initialize w/o having a reference to it; updated db.h to include all proper API calls
 * Some rcsid for db_* were wrong; fixed
 *
 * Revision 1.2  2004/02/25 06:14:00  ryalth
 * Changed to allow \n and \r in strings,  and preserve across flush.  Modified read_string, read_line and write_line
 *
 * Revision 1.1  2004/01/31 07:38:08  luke-jr
 * Merged Database Format Selection support
 *
 */
