/******************************************************************************
  Copyright (c) 1992, 1995, 1996 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-ctype.h"
#include "my-string.h"

#include "bf_register.h"
#include "config.h"
#include "exceptions.h"
#include "functions.h"
#include "list.h"
#include "log.h"
#include "md5.h"
#include "options.h"
#include "pattern.h"
#include "random.h"
#include "ref_count.h"
#include "streams.h"
#include "storage.h"
#include "structures.h"
#include "unparse.h"
#include "utils.h"
#include "utf8.h"

Var
new_list(int size)
{
    Var new;

    if (size == 0) {
	static Var emptylist;

	if (emptylist.v.list == 0) {
	    emptylist.type = TYPE_LIST;
	    emptylist.v.list = mymalloc(1 * sizeof(Var), M_LIST);
	    emptylist.v.list[0].type = TYPE_INT;
	    emptylist.v.list[0].v.num = 0;
	}
	/* give the lucky winner a reference */
	addref(emptylist.v.list);
	return emptylist;
    }
    new.type = TYPE_LIST;
    new.v.list = (Var *) mymalloc((size + 1) * sizeof(Var), M_LIST);
    new.v.list[0].type = TYPE_INT;
    new.v.list[0].v.num = size;
    return new;
}

Var
make_list(int size, Var elt)
{
	Var ret = new_list(size);
	int i;
	
	if(size == 0)
		return ret;
	
	for(i = 1; i <= size; i++)
		ret.v.list[i] = var_dup(elt);
	
	return ret;
}

Var
setadd(Var list, Var value)
{
    if (ismember(value, list, 0)) {
	free_var(value);
	return list;
    }
    return listappend(list, value);
}

Var
setremove(Var list, Var value)
{
    int i;

    if ((i = ismember(value, list, 0)) != 0) {
	return listdelete(list, i);
    } else {
	return list;
    }
}

int
ismember(Var lhs, Var rhs, int case_matters)
{
    int i;

    for (i = 1; i <= rhs.v.list[0].v.num; i++) {
	if (equality(lhs, rhs.v.list[i], case_matters)) {
	    return i;
	}
    }
    return 0;
}

Var
listset(Var list, Var value, int pos)
{
    free_var(list.v.list[pos]);
    list.v.list[pos] = value;
    return list;
}

static Var
doinsert(Var list, Var value, int pos)
{
    Var new;
    int i;
    int size = list.v.list[0].v.num + 1;

    if (var_refcount(list) == 1 && pos == size) {
	list.v.list = (Var *) myrealloc(list.v.list, (size + 1) * sizeof(Var), M_LIST);
	list.v.list[0].v.num = size;
	list.v.list[pos] = value;
	return list;
    }
    new = new_list(size);
    for (i = 1; i < pos; i++)
	new.v.list[i] = var_ref(list.v.list[i]);
    new.v.list[pos] = value;
    for (i = pos; i <= list.v.list[0].v.num; i++)
	new.v.list[i + 1] = var_ref(list.v.list[i]);
    free_var(list);
    return new;
}

Var
listinsert(Var list, Var value, int pos)
{
    if (pos <= 0)
	pos = 1;
    else if (pos > list.v.list[0].v.num)
	pos = list.v.list[0].v.num + 1;
    return doinsert(list, value, pos);
}

Var
listappend(Var list, Var value)
{
    return doinsert(list, value, list.v.list[0].v.num + 1);
}

Var
listdelete(Var list, int pos)
{
    Var new;
    int i;

    new = new_list(list.v.list[0].v.num - 1);
    for (i = 1; i < pos; i++) {
	new.v.list[i] = var_ref(list.v.list[i]);
    }
    for (i = pos + 1; i <= list.v.list[0].v.num; i++)
	new.v.list[i - 1] = var_ref(list.v.list[i]);
    free_var(list);		/* free old list */
    return new;
}

Var
listslice(Var list, Var ind)
{
	/* slice(alist[,index]) returns a list of the index-th elements of the elements of alist, e.g., 
	 *     slice({{"z",1},{"y",2},{"x",5}},2) => {1,2,5}.
	 * index defaults to 1 and may also be a nonempty list, e.g., 
	 *     slice({{"z",1,3},{"y",2,4}},{2,1}) => {{1,"z"},{2,"y"}}
	 */
	Var slice, s = zero;
	int y, z;

#define RETURN_ERROR(e) {	\
	free_var(ind);			\
	free_var(slice);		\
	free_var(s);			\
	slice.type = TYPE_ERR;	\
	slice.v.err = e;		\
	return slice; }
	
	// Note: we use make_list so that we can safely free_var without worrying about uninitialized portions of the list.
	
	if (ind.type != TYPE_LIST && ind.type != TYPE_INT && list.type != TYPE_LIST)
		RETURN_ERROR(E_TYPE)
	slice = make_list(list.v.list[0].v.num, zero);
	if (ind.type == TYPE_LIST)
		for (z = 1; z <= list.v.list[0].v.num; ++z) {
			s = make_list(ind.v.list[0].v.num, zero);
			for (y = 1; y <= ind.v.list[0].v.num; ++y)
				if (list.v.list[z].type == TYPE_LIST)
					if (list.v.list[z].v.list[0].v.num < ind.v.list[y].v.num)
						RETURN_ERROR(E_RANGE)
					else
						s.v.list[y] = var_ref(list.v.list[z].v.list[ind.v.list[y].v.num]);
				else if (list.v.list[z].type == TYPE_STR)
					if (strlen(list.v.list[z].v.str) < ind.v.list[y].v.num)
						RETURN_ERROR(E_RANGE)
					else
						s.v.list[y] = strget(list.v.list[z], ind.v.list[y]);
				else
					RETURN_ERROR(E_TYPE)
			slice.v.list[z] = s;
		}
	else	// TYPE_INT
		for (z = 1; z <= list.v.list[0].v.num; ++z)
			if (list.v.list[z].type == TYPE_LIST)
				if (list.v.list[z].v.list[0].v.num < ind.v.num)
					RETURN_ERROR(E_RANGE)
				else
					slice.v.list[z] = var_ref(list.v.list[z].v.list[ind.v.num]);
			else if (list.v.list[z].type == TYPE_STR)
				if (strlen(list.v.list[z].v.str) < ind.v.num)
					RETURN_ERROR(E_RANGE)
				else
					slice.v.list[z] = strget(list.v.list[z], ind);
			else
				RETURN_ERROR(E_TYPE);
#undef RETURN_ERROR
	return slice;
}

Var
remove_duplicates(Var list)
{
	Var	r;
	int	i;

	r = new_list(0);
	for (i = 1; i <= list.v.list[1].v.list[0].v.num; i++)
		r = setadd(r, var_ref(list.v.list[1].v.list[i]));

	free_var(list);
	return r;
}

int
find_insert(Var lst, Var key)
{
/* find_insert(sortedlist,key) => index of first element in sortedlist
   greater than key.  sortedlist is assumed to be sorted in increasing
   order and the number returned is anywhere from 1 to length(sortedlist)+1,
   inclusive. */

/* returns 0 if an error occurs */

	Var	compare;
	int	r = lst.v.list[0].v.num, l=1, i;

	while(r >= l) {
		compare = value_compare(var_ref(key), var_ref(lst.v.list[i = ((r + l) / 2)] ) );
		if(compare.type == TYPE_ERR) {
			free_var(compare);
			return 0;
		}
		if(compare.v.num < 0) {
			r = i - 1;
		} else {
			l = i + 1;
		}
	}
	return l;
}

int
find_insert_always(Var lst, Var key)
{	/* Same as find_insert, but will ignore any errors and implements a
	 * binary search algorithm
	 */
	int r = lst.v.list[0].v.num, l, i;
	Var v;
	
	l = 1;
	while (r >= l) {
		if ((v = value_compare(lst.v.list[i = (r + l) / 2], key)).type == TYPE_INT && v.v.num > 0)
			r = i - 1;
		else
			l = i + 1;
		free_var(v);
	}
	return l;
}

int
list_iassoc(Var vtarget, Var vlist, int vindex)
{
  int i;

  for (i = 1; i <= vlist.v.list[0].v.num; i++) {
    if (vlist.v.list[i].type == TYPE_LIST &&
        vlist.v.list[i].v.list[0].v.num >= vindex &&
        equality(vlist.v.list[i].v.list[vindex], vtarget, 0)) {

      return i;
    }
  }
  return 0;
}

Var
list_assoc(Var vtarget, Var vlist, int vindex)
{
  int i;

  for (i = 1; i <= vlist.v.list[0].v.num; i++) {
    if (vlist.v.list[i].type == TYPE_LIST &&
        vlist.v.list[i].v.list[0].v.num >= vindex &&
        equality(vlist.v.list[i].v.list[vindex], vtarget, 0)) {

      return var_dup(vlist.v.list[i]);
    }
  }
  return new_list(0);
}

Var
listconcat(Var first, Var second)
{
    int lsecond = second.v.list[0].v.num;
    int lfirst = first.v.list[0].v.num;
    Var new;
    int i;

    new = new_list(lsecond + lfirst);
    for (i = 1; i <= lfirst; i++)
	new.v.list[i] = var_ref(first.v.list[i]);
    for (i = 1; i <= lsecond; i++)
	new.v.list[i + lfirst] = var_ref(second.v.list[i]);
    free_var(first);
    free_var(second);

    return new;
}

Var
list_key_sort(Var lst, Var unsorted_keys)
{
	/* sort(list[,keys]) => sorts keys (assumed to be all numbers or strings) and
	 * returns list with the corresponding permutation applied to it.  keys
	 * defaults to the list itself.
	 * example: sort({x1,x3,x2},{1,3,2}) => {x1,x2,x3}
	 */
	int use_sorted_lst = !equality(lst, unsorted_keys, 0);
	Var sorted_lst, sorted_keys, e;
	int i, l;
	
	if (lst.v.list[0].v.num < unsorted_keys.v.list[0].v.num) {
		e.type = TYPE_ERR;
		e.v.err = E_RANGE;
		free_var(lst);
		free_var(unsorted_keys);
		return e;
	}
	
	sorted_keys = new_list(0);
	if (use_sorted_lst)
		sorted_lst = new_list(0);
	for (i = 1; i <= unsorted_keys.v.list[0].v.num; ++i) {
		e = unsorted_keys.v.list[i];
		l = find_insert_always(sorted_keys, e);
		sorted_keys = listinsert(sorted_keys, var_ref(e), l);
		if (use_sorted_lst)
			sorted_lst = listinsert(sorted_lst, var_ref(lst.v.list[sorted_keys.v.list[0].v.num]), l);
	}
	free_var(lst);
	free_var(unsorted_keys);
	if (use_sorted_lst) {
		free_var(sorted_keys);
		sorted_keys = sorted_lst;
	}
	return sorted_keys;
}

Var
list_sort(Var l)
{
	return list_key_sort(l, l);
}

Var
listrangeset(Var base, int from, int to, Var value)
{
    /* base and value are free'd */
    int index, offset = 0;
    int val_len = value.v.list[0].v.num;
    int base_len = base.v.list[0].v.num;
    int lenleft = (from > 1) ? from - 1 : 0;
    int lenmiddle = val_len;
    int lenright = (base_len > to) ? base_len - to : 0;
    int newsize = lenleft + lenmiddle + lenright;
    Var ans;

    ans = new_list(newsize);
    for (index = 1; index <= lenleft; index++)
	ans.v.list[++offset] = var_ref(base.v.list[index]);
    for (index = 1; index <= lenmiddle; index++)
	ans.v.list[++offset] = var_ref(value.v.list[index]);
    for (index = 1; index <= lenright; index++)
	ans.v.list[++offset] = var_ref(base.v.list[to + index]);
    free_var(base);
    free_var(value);
    return ans;
}

Var
sublist(Var list, int lower, int upper)
{
    if (lower > upper) {
	free_var(list);
	return new_list(0);
    } else {
	Var r;
	int i;

	r = new_list(upper - lower + 1);
	for (i = lower; i <= upper; i++)
	    r.v.list[i - lower + 1] = var_ref(list.v.list[i]);
	free_var(list);
	return r;
    }
}

static const char *
list2str(Var * args)
{
    static Stream *str = 0;
    int i;

    if (!str)
	str = new_stream(100);

    for (i = 1; i <= args[0].v.num; i++) {
	switch (args[i].type) {
	case TYPE_INT:
	    stream_printf(str, "%d", args[i].v.num);
	    break;
	case TYPE_OBJ:
	    stream_printf(str, "#%d", args[i].v.obj);
	    break;
	case TYPE_STR:
	    stream_add_string(str, args[i].v.str);
	    break;
	case TYPE_ERR:
	    stream_add_string(str, unparse_error(args[i].v.err));
	    break;
	case TYPE_FLOAT:
	    stream_printf(str, "%g", *args[i].v.fnum);
	    break;
	case TYPE_LIST:
	    stream_add_string(str, "{list}");
	    break;
	default:
	    panic("LIST2STR: Impossible var type.\n");
	}
    }

    return reset_stream(str);
}

const char *
value2str(Var value)
{
    Var list;
    const char *str;

    list = new_list(1);
    list.v.list[1] = var_ref(value);
    str = list2str(list.v.list);
    free_var(list);
    return str;
}

static void
print_to_stream(Var v, Stream * s)
{
    switch (v.type) {
    case TYPE_INT:
	stream_printf(s, "%d", v.v.num);
	break;
    case TYPE_OBJ:
	stream_printf(s, "#%d", v.v.obj);
	break;
    case TYPE_ERR:
	stream_add_string(s, error_name(v.v.num));
	break;
    case TYPE_FLOAT:
	stream_printf(s, "%g", *v.v.fnum);
	break;
    case TYPE_STR:
		{
			const char *str = string_quote(v.v.str);
			
			stream_add_string(s, str);
			free_str(str);
		}
	break;
    case TYPE_LIST:
	{
	    const char *sep = "";
	    int len, i;

	    stream_add_char(s, '{');
	    len = v.v.list[0].v.num;
	    for (i = 1; i <= len; i++) {
		stream_add_string(s, sep);
		sep = ", ";
		print_to_stream(v.v.list[i], s);
	    }
	    stream_add_char(s, '}');
	}
	break;
    default:
	errlog("PRINT_TO_STREAM: Unknown Var type = %d\n", v.type);
	stream_add_string(s, ">>Unknown value<<");
    }
}

const char *
value_to_literal(Var v)
{
    static Stream *s = 0;

    if (!s)
	s = new_stream(100);

    print_to_stream(v, s);

    return reset_stream(s);
}

Var
strrangeset(Var base, int from, int to, Var value)
{
    Var ans;

    ans.type = TYPE_STR;
    ans.v.str = utf8_strrangeset(base.v.str, from, to, value.v.str);

    return ans;
}

char *
strsetoctets(const char *base, const char *value, int from, int to)
{
	int index, offset = 0;
	int val_len = strlen(value), base_len = strlen(base);
	int lenleft = (from > 1) ? from - 1 : 0;
	int lenmiddle = val_len;
	int lenright = (base_len > to) ? base_len - to : 0;
	int newsize = lenleft + lenmiddle + lenright;

	char *s;

	s = mymalloc(sizeof(char) * (newsize + 1), M_STRING);

	for (index = 0; index < lenleft; index++)
		s[offset++] = base[index];
	for (index = 0; index < lenmiddle; index++)
		s[offset++] = value[index];
	for (index = 0; index < lenright; index++)
		s[offset++] = base[index + to];
	s[offset] = '\0';
	return s;
}

Var
substr(Var str, int lower, int upper)
{
    Var r;

    r.type = TYPE_STR;
	r.v.str = utf8_substr(str.v.str, lower, upper);
    free_var(str);
    return r;
}

char *
stroctets(const char *str, int from, int to)
{
	char *s;
	if (from > to || from > strlen(str) || from < 1)
		s = str_dup("");
	else {
		int loop, index = 0;
		s = mymalloc(to - from + 2, M_STRING);

		for (loop = from - 1; loop < to; loop++)
			s[index++] = str[loop];
		s[index] = '\0';
	}
	
	return s;
}

Var
strget(Var str, Var i)
{
    Var r;

    r.type = TYPE_STR;
    r.v.str = utf8_index(str.v.str, i.v.num);

    return r;
}


Var
char_list(const char *s)
{
	Var r;
	int l = utf8_strlen(s), i;
	const char *ss = s;
	
	r = new_list(l);
	for (i = 1; i <= l; ++i) {
		r.v.list[i].type = TYPE_STR;
		r.v.list[i].v.str = utf8_index(ss, 1);
		ss += strlen(r.v.list[i].v.str);
	}
	
	return r;
}

Var
strexplode(const char *subject, const char *delim)
{
	/* explode(subject [, delim])
	 * Return a list of those substrings of subject separated by runs of delim[1].
	 * delim defaults to space.
	 */
	Var item, parts = new_list(0);
	const char *s = subject, *d = utf8_index(delim, 1);
	int i;
	
	if (delim[0] == '\0')
		return char_list(subject);

	item.type = TYPE_STR;
	while (s) {
		if ((i = utf8_strindex(s, d, 0)) > 1) {
			item.v.str = utf8_substr(s, 1, i - 1);
			parts = listappend(parts, item);
		} else if (i == 0) {
			item.v.str = str_dup(s);
			parts = listappend(parts, item);
			break;
		}
		s += i + strlen(utf8_index(s, i)) - 1;
	}
	
	return parts;
}

int
is_options_list(Var options)
{
  Var opt;
  int len, i;

  if (options.type != TYPE_LIST) {
    return 0;
  }

  len = options.v.list[0].v.num;
  for (i = 1; i <= len; i++) {
    opt = options.v.list[i];
    if (opt.type != TYPE_LIST) {
      return 0;
    } else if (opt.v.list[0].v.num != 2) {
      return 0;
    } else if (opt.v.list[1].type != TYPE_STR) {
      return 0;
    }
  }

  return 1;
}

int
options_list_value(Var options, const char *name, Var *value)
{
  Var opt;
  int len, i;

  len = options.v.list[0].v.num;
  for (i = 1; i <= len; i++) {
    opt = options.v.list[i];
    if (!mystrcasecmp(name, opt.v.list[1].v.str)) {
      *value = opt.v.list[2];
      return 1;
    }
  }

  return 0;
}


/**** built in functions ****/

static package
bf_length(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    switch (arglist.v.list[1].type) {
    case TYPE_LIST:
	r.type = TYPE_INT;
	r.v.num = arglist.v.list[1].v.list[0].v.num;
	break;
    case TYPE_STR:
	r.type = TYPE_INT;
	r.v.num = utf8_strlen(arglist.v.list[1].v.str);
	break;
    default:
	free_var(arglist);
	return make_error_pack(E_TYPE);
	break;
    }

    free_var(arglist);
    return make_var_pack(r);
}


static package
bf_slice(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var ret, c;
	enum error e;
	
	if (!is_wizard(progr))
		return make_raise_pack(E_PERM, "Temporarily requires wizperms (Security)", zero);
	
	if (arglist.v.list[0].v.num == 2)
		if (arglist.v.list[2].type == TYPE_INT || arglist.v.list[2].type == TYPE_LIST)
			c = var_dup(arglist.v.list[2]);
		else {
			free_var(arglist);
			return make_error_pack(E_TYPE);
		}
	else {
		c.type = TYPE_INT;
		c.v.num = 1;
	}
	
	ret = listslice(arglist.v.list[1], c);
	
	free_var(arglist);
	if(ret.type == TYPE_ERR) {
		e = ret.v.err;
		free_var(ret);
		return make_error_pack(e);
	} else
		return make_var_pack(ret);
}


static package
bf_remove_duplicates(Var arglist, Byte next, void *vdata, Objid progr)
{
	if (!is_wizard(progr))
		return make_raise_pack(E_PERM, "Temporarily requires wizperms (Security)", zero);
	
	return make_var_pack(remove_duplicates(arglist));
}


static package
bf_octets(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var r;
	
	r.type = TYPE_INT;
	r.v.num = strlen(arglist.v.list[1].v.str);
	free_var(arglist);
	return make_var_pack(r);
}

static package
bf_getoctets(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var base = arglist.v.list[1], rv;
	int to, from = arglist.v.list[2].v.num;
	
	rv.type = TYPE_STR;
	
	if (arglist.v.list[0].v.num > 2)
		to = arglist.v.list[3].v.num;
	else
		to = strlen(base.v.str);
	
	if (from < 1) {
		free_var(arglist);
		return make_error_pack(E_RANGE);
	} else
		rv.v.str = stroctets(base.v.str, from, to);
		
	free_var(arglist);
	return make_var_pack(rv);
}

static package
bf_getoctet(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var base = arglist.v.list[1], rv;
	int i = arglist.v.list[2].v.num;
	
	rv.type = TYPE_STR;
	
	if (i < 1 || i > strlen(base.v.str)) {
		free_var(arglist);
		return make_error_pack(E_RANGE);
	} else
		rv.v.str = stroctets(base.v.str, i, i);
		
	free_var(arglist);
	return make_var_pack(rv);
}

static package
bf_setoctets(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var base = arglist.v.list[1], value = arglist.v.list[2], rv;
	int from = arglist.v.list[3].v.num, to;
	
	rv.type = TYPE_STR;
	
	if (arglist.v.list[0].v.num > 3)
		to = arglist.v.list[4].v.num;
	else
		to = strlen(base.v.str);
	
	if (from < 1 || from > strlen(base.v.str) + 1) {
		free_var(arglist);
		return make_error_pack(E_RANGE);
	} else
		rv.v.str = strsetoctets(base.v.str, value.v.str, from, to);
	
	free_var(arglist);
	return make_var_pack(rv);
}

static package
bf_setoctet(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var base = arglist.v.list[1], value = arglist.v.list[2], rv;
	int i = arglist.v.list[3].v.num;
	
	rv.type = TYPE_STR;
	
	if (i < 1 || i > strlen(base.v.str) + 1) {
		free_var(arglist);
		return make_error_pack(E_RANGE);
	} else
		rv.v.str = strsetoctets(base.v.str, value.v.str, i, i);
	
	free_var(arglist);
	return make_var_pack(rv);
}

static package
bf_setadd(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;

    r = setadd(var_ref(arglist.v.list[1]), var_ref(arglist.v.list[2]));
    free_var(arglist);
    return make_var_pack(r);
}


static package
bf_setremove(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;

    r = setremove(var_ref(arglist.v.list[1]), arglist.v.list[2]);
    free_var(arglist);
    return make_var_pack(r);
}


static package
bf_enlist(Var arglist, Byte next, void *vdata, Objid progr)
{	/* Enlist - Make args[1] into a list, if it isn't. */
	Var r;

	if(arglist.v.list[1].type != TYPE_LIST) {
		r = new_list(1);
		r.v.list[1] = var_ref(arglist.v.list[1]);
	} else {
		r = var_ref(arglist.v.list[1]);
	}
	free_var(arglist);
	return make_var_pack(r);
}


static package
bf_make_list(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var ret, elt;
	int n = arglist.v.list[1].v.num;

	if (!is_wizard(progr))
		return make_raise_pack(E_PERM, "Temporarily requires wizperms (Security)", zero);
	
	if(n < 0) {
		free_var(arglist);
		make_error_pack(E_INVARG);
	}
	
	elt = (arglist.v.list[0].v.num == 2) ? var_dup(arglist.v.list[2]) : zero;
	ret = make_list(n, elt);
	
	free_var(elt);
	free_var(arglist);
	return make_var_pack(ret);
}


static package
bf_sort(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var r = list_key_sort(var_ref(arglist.v.list[1]), var_ref(arglist.v.list[(arglist.v.list[0].v.num > 1) ? 2 : 1]));
	
	if (!is_wizard(progr))
		return make_raise_pack(E_PERM, "Temporarily requires wizperms (Security)", zero);
	
	free_var(arglist);
	if (r.type == TYPE_ERR) {
		int n = r.v.err;
		free_var(r);
		return make_error_pack(n);
	} else
		return make_var_pack(r);
}


static package
bf_listappend(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    if (arglist.v.list[0].v.num == 2)
	r = listappend(var_ref(arglist.v.list[1]), var_ref(arglist.v.list[2]));
    else
	r = listinsert(var_ref(arglist.v.list[1]), var_ref(arglist.v.list[2]),
		       arglist.v.list[3].v.num + 1);
    free_var(arglist);
    return make_var_pack(r);
}


static package
bf_listinsert(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    if (arglist.v.list[0].v.num == 2)
	r = listinsert(var_ref(arglist.v.list[1]), var_ref(arglist.v.list[2]), 1);
    else
	r = listinsert(var_ref(arglist.v.list[1]),
		    var_ref(arglist.v.list[2]), arglist.v.list[3].v.num);
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_listdelete(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    if (arglist.v.list[2].v.num <= 0
	|| arglist.v.list[2].v.num > arglist.v.list[1].v.list[0].v.num) {
	free_var(arglist);
	return make_error_pack(E_RANGE);
    } else {
	r = listdelete(var_ref(arglist.v.list[1]), arglist.v.list[2].v.num);
    }
    free_var(arglist);
    return make_var_pack(r);
}


static package
bf_listset(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    if (arglist.v.list[3].v.num <= 0
	|| arglist.v.list[3].v.num > arglist.v.list[1].v.list[0].v.num) {
	free_var(arglist);
	return make_error_pack(E_RANGE);
    } else {
	r = listset(var_dup(arglist.v.list[1]),
		    var_ref(arglist.v.list[2]), arglist.v.list[3].v.num);
    }
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_equal(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;

    r.type = TYPE_INT;
    r.v.num = equality(arglist.v.list[1], arglist.v.list[2], 1);
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_is_member(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;

    r.type = TYPE_INT;
    r.v.num = ismember(arglist.v.list[1], arglist.v.list[2], 1);
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_iassoc(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var r;
	int index = 1;
	
	if (!is_wizard(progr))
		return make_raise_pack(E_PERM, "Temporarily requires wizperms (Security)", zero);
	
	r.type = TYPE_INT;
	if (arglist.v.list[0].v.num == 3)
		index = arglist.v.list[3].v.num;
	
	if (index < 1) {
		free_var(arglist);
		return make_error_pack(E_RANGE);
	}
	
	r.v.num = list_iassoc(arglist.v.list[1], arglist.v.list[2], index);
	
	free_var(arglist);
	return make_var_pack(r);
}

static package
bf_assoc(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var r;
	int index = 1;
	
	if (!is_wizard(progr))
		return make_raise_pack(E_PERM, "Temporarily requires wizperms (Security)", zero);
	
	if (arglist.v.list[0].v.num == 3)
		index = arglist.v.list[3].v.num;
	
	if (index < 1) {
		free_var(arglist);
		return make_error_pack(E_RANGE);
	}
	
	r = list_assoc(arglist.v.list[1], arglist.v.list[2], index);
	
	free_var(arglist);
	return make_var_pack(r);
}

static package
bf_random_of(Var arglist, Byte next, void *vdata, Objid progr)
{	/* Random_Of - Random item of a list. */
	Var 	r;
	int	n;

	n = arglist.v.list[1].v.list[0].v.num;

	if(!n) {
		free_var(arglist);
		return make_error_pack(E_INVARG);
	}

	n = RANDOM() % n + 1;
	r = var_ref(arglist.v.list[1].v.list[n]);

	free_var(arglist);
        return make_var_pack(r);
}

static package
bf_strsub(Var arglist, Byte next, void *vdata, Objid progr)
{				/* (source, what, with [, case-matters]) */
    Var r;
    int case_matters = 0;

    if (arglist.v.list[0].v.num == 4)
	case_matters = is_true(arglist.v.list[4]);
    if (arglist.v.list[2].v.str[0] == '\0') {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	r.type = TYPE_STR;
	r.v.str = str_dup(strsub(arglist.v.list[1].v.str,
				 arglist.v.list[2].v.str,
				 arglist.v.list[3].v.str, case_matters));

	free_var(arglist);
	return make_var_pack(r);
    }
}

static package
bf_crypt(Var arglist, Byte next, void *vdata, Objid progr)
{				/* (string, [salt]) */
    Var r;

#if HAVE_CRYPT
    char salt[3];
    const char *saltp;
    static char saltstuff[] =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
    extern const char *crypt(const char *, const char *);

    if (arglist.v.list[0].v.num == 1 || strlen(arglist.v.list[2].v.str) < 2) {
	/* provide a random 2-letter salt, works with old and new crypts */
	salt[0] = saltstuff[RANDOM() % (int) strlen(saltstuff)];
	salt[1] = saltstuff[RANDOM() % (int) strlen(saltstuff)];
	salt[2] = '\0';
	saltp = salt;
    } else {
	/* return the entire crypted password in the salt, this works
	 * for all crypt versions */
	saltp = arglist.v.list[2].v.str;
    }
    r.type = TYPE_STR;
    r.v.str = str_dup(crypt(arglist.v.list[1].v.str, saltp));
#else				/* !HAVE_CRYPT */
    r.type = TYPE_STR;
    r.v.str = str_ref(arglist.v.list[1].v.str);
#endif

    free_var(arglist);
    return make_var_pack(r);
}

static int
signum(int x)
{
    return x < 0 ? -1 : (x > 0 ? 1 : 0);
}

static package
bf_strcmp(Var arglist, Byte next, void *vdata, Objid progr)
{				/* (string1, string2) */
    Var r;

    r.type = TYPE_INT;
    r.v.num = signum(strcmp(arglist.v.list[1].v.str, arglist.v.list[2].v.str));
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_index(Var arglist, Byte next, void *vdata, Objid progr)
{				/* (source, what [, case-matters]) */
    Var r;
    int case_matters = 0;

    if (arglist.v.list[0].v.num == 3)
	case_matters = is_true(arglist.v.list[3]);
    r.type = TYPE_INT;
    r.v.num = utf8_strindex(arglist.v.list[1].v.str, arglist.v.list[2].v.str,
		       case_matters);

    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_rindex(Var arglist, Byte next, void *vdata, Objid progr)
{				/* (source, what [, case-matters]) */
    Var r;

    int case_matters = 0;

    if (arglist.v.list[0].v.num == 3)
	case_matters = is_true(arglist.v.list[3]);
    r.type = TYPE_INT;
    r.v.num = utf8_strrindex(arglist.v.list[1].v.str, arglist.v.list[2].v.str,
			case_matters);

    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_tostr(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    r.type = TYPE_STR;
    r.v.str = str_dup(list2str(arglist.v.list));
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_toliteral(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;

    r.type = TYPE_STR;
    r.v.str = str_dup(value_to_literal(arglist.v.list[1]));
    free_var(arglist);
    return make_var_pack(r);
}

struct pat_cache_entry {
    char *string;
    int case_matters;
    Pattern pattern;
    struct pat_cache_entry *next;
};

static struct pat_cache_entry *pat_cache;
static struct pat_cache_entry pat_cache_entries[PATTERN_CACHE_SIZE];

static void
setup_pattern_cache()
{
    int i;

    for (i = 0; i < PATTERN_CACHE_SIZE; i++) {
	pat_cache_entries[i].string = 0;
	pat_cache_entries[i].pattern.ptr = 0;
    }

    for (i = 0; i < PATTERN_CACHE_SIZE - 1; i++)
	pat_cache_entries[i].next = &(pat_cache_entries[i + 1]);
    pat_cache_entries[PATTERN_CACHE_SIZE - 1].next = 0;

    pat_cache = &(pat_cache_entries[0]);
}

static Pattern
get_pattern(const char *string, int case_matters)
{
    struct pat_cache_entry *entry, **entry_ptr;

    entry = pat_cache;
    entry_ptr = &pat_cache;

    while (1) {
	if (entry->string && !strcmp(string, entry->string)
	    && case_matters == entry->case_matters) {
	    /* A cache hit; move this entry to the front of the cache. */
	    break;
	} else if (!entry->next) {
	    /* A cache miss; this is the last entry in the cache, so reuse that
	     * one for this pattern, moving it to the front of the cache iff
	     * the compilation succeeds.
	     */
	    if (entry->string) {
		free_str(entry->string);
		free_pattern(entry->pattern);
	    }
	    entry->pattern = new_pattern(string, case_matters);
	    entry->case_matters = case_matters;
	    if (!entry->pattern.ptr)
		entry->string = 0;
	    else
		entry->string = str_dup(string);
	    break;
	} else {
	    /* not done searching the cache... */
	    entry_ptr = &(entry->next);
	    entry = entry->next;
	}
    }

    *entry_ptr = entry->next;
    entry->next = pat_cache;
    pat_cache = entry;
    return entry->pattern;
}

Var
do_match(Var arglist, int reverse)
{
    const char *subject, *pattern;
    int i;
    Pattern pat;
    Var ans;
    Match_Indices regs[10];

    subject = arglist.v.list[1].v.str;
    pattern = arglist.v.list[2].v.str;
    pat = get_pattern(pattern, (arglist.v.list[0].v.num == 3
				&& is_true(arglist.v.list[3])));

    if (!pat.ptr) {
	ans.type = TYPE_ERR;
	ans.v.err = E_INVARG;
    } else
	switch (match_pattern(pat, subject, regs, reverse)) {
	case MATCH_SUCCEEDED:
	    ans = new_list(4);
	    ans.v.list[1].type = TYPE_INT;
	    ans.v.list[2].type = TYPE_INT;
	    ans.v.list[4].type = TYPE_STR;
	    ans.v.list[1].v.num = utf8_convert_index(regs[0].start, subject);
	    ans.v.list[2].v.num = (regs[0].end == 0 ? 0 : utf8_convert_index(regs[0].end, subject));
	    ans.v.list[3] = new_list(9);
	    ans.v.list[4].v.str = str_ref(subject);
	    for (i = 1; i <= 9; i++) {
		ans.v.list[3].v.list[i] = new_list(2);
		ans.v.list[3].v.list[i].v.list[1].type = TYPE_INT;
		ans.v.list[3].v.list[i].v.list[1].v.num = (regs[i].start == 0 ? 0 : utf8_convert_index(regs[i].start, subject));
		ans.v.list[3].v.list[i].v.list[2].type = TYPE_INT;
		ans.v.list[3].v.list[i].v.list[2].v.num = (regs[i].end <= 0 ? regs[i].end : utf8_convert_index(regs[i].end, subject));
	    }
	    break;
	case MATCH_FAILED:
	    ans = new_list(0);
	    break;
	case MATCH_ABORTED:
	    ans.type = TYPE_ERR;
	    ans.v.err = E_QUOTA;
	    break;
	}

    return ans;
}

static package
bf_match(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var ans;

    ans = do_match(arglist, 0);
    free_var(arglist);
    if (ans.type == TYPE_ERR)
	return make_error_pack(ans.v.err);
    else
	return make_var_pack(ans);
}

static package
bf_rmatch(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var ans;

    ans = do_match(arglist, 1);
    free_var(arglist);
    if (ans.type == TYPE_ERR)
	return make_error_pack(ans.v.err);
    else
	return make_var_pack(ans);
}

int
invalid_pair(int num1, int num2, int max)
{
    if ((num1 == 0 && num2 == -1)
	|| (num1 > 0 && num2 >= num1 - 1 && num2 <= max))
	return 0;
    else
	return 1;
}

int
check_subs_list(Var subs)
{
    const char *subj;
    int subj_length, loop;

    if (subs.type != TYPE_LIST || subs.v.list[0].v.num != 4
	|| subs.v.list[1].type != TYPE_INT
	|| subs.v.list[2].type != TYPE_INT
	|| subs.v.list[3].type != TYPE_LIST
	|| subs.v.list[3].v.list[0].v.num != 9
	|| subs.v.list[4].type != TYPE_STR)
	return 1;
    subj = subs.v.list[4].v.str;
    subj_length = strlen(subj);
    if (invalid_pair(subs.v.list[1].v.num, subs.v.list[2].v.num,
		     subj_length))
	return 1;

    for (loop = 1; loop <= 9; loop++) {
	Var pair;
	pair = subs.v.list[3].v.list[loop];
	if (pair.type != TYPE_LIST
	    || pair.v.list[0].v.num != 2
	    || pair.v.list[1].type != TYPE_INT
	    || pair.v.list[2].type != TYPE_INT
	    || invalid_pair(pair.v.list[1].v.num, pair.v.list[2].v.num,
			    subj_length))
	    return 1;
    }
    return 0;
}

static package
bf_substitute(Var arglist, Byte next, void *vdata, Objid progr)
{
    int template_length, subject_length;
    const char *template, *subject;
    Var subs, ans;
    int invarg = 0;
    Stream *s;
    char c = '\0';

    template = arglist.v.list[1].v.str;
    template_length = strlen(template);
    subs = arglist.v.list[2];

    if (check_subs_list(subs)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    }
    subject = subs.v.list[4].v.str;
    subject_length = strlen(subject);

    s = new_stream(template_length);
    ans.type = TYPE_STR;
    while ((c = *(template++)) != '\0') {
	switch (c) {
	case '%':
	    {
		Var pair;
		int start = 0, end = 0;
		c = *(template++);
		if (c == '%')
		    stream_add_char(s, '%');
		else {
		    if (c >= '1' && c <= '9') {
			pair = subs.v.list[3].v.list[c - '0'];
			start = pair.v.list[1].v.num - 1;
			end = pair.v.list[2].v.num - 1;
		    } else if (c == '0') {
			start = subs.v.list[1].v.num - 1;
			end = subs.v.list[2].v.num - 1;
		    } else
			invarg = 1;
		    if (!invarg) {
			int where;
			for (where = start; where <= end; where++)
			    stream_add_char(s, subject[where]);
		    }
		}
		break;
	    }
	default:
	    stream_add_char(s, c);
	}
	if (invarg)
	    break;
    }

    free_var(arglist);
    if (!invarg)
	ans.v.str = str_dup(reset_stream(s));
    free_stream(s);
    if (invarg)
	return make_error_pack(E_INVARG);
    else
	return make_var_pack(ans);
}

static package
bf_value_bytes(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;

    r.type = TYPE_INT;
    r.v.num = value_bytes(arglist.v.list[1]);
    free_var(arglist);
    return make_var_pack(r);
}

static const char *
hash_bytes(const char *input, int length)
{
    md5ctx_t context;
    uint8 result[16];
    int i;
    const char digits[] = "0123456789ABCDEF";
    char *hex = str_dup("12345678901234567890123456789012");
    const char *answer = hex;

    md5_Init(&context);
    md5_Update(&context, (uint8 *) input, length);
    md5_Final(&context, result);
    for (i = 0; i < 16; i++) {
	*hex++ = digits[result[i] >> 4];
	*hex++ = digits[result[i] & 0xF];
    }
    return answer;
}

static package
bf_binary_hash(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    int length;
    const char *bytes = binary_to_raw_bytes(arglist.v.list[1].v.str, &length);

    free_var(arglist);
    if (!bytes)
	return make_error_pack(E_INVARG);
    r.type = TYPE_STR;
    r.v.str = hash_bytes(bytes, length);
    return make_var_pack(r);
}

static package
bf_string_hash(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    const char *str = arglist.v.list[1].v.str;

    r.type = TYPE_STR;
    r.v.str = hash_bytes(str, strlen(str));
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_value_hash(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    const char *lit = value_to_literal(arglist.v.list[1]);

    r.type = TYPE_STR;
    r.v.str = hash_bytes(lit, strlen(lit));
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_decode_binary(Var arglist, Byte next, void *vdata, Objid progr)
{
    int length;
    const char *bytes = binary_to_raw_bytes(arglist.v.list[1].v.str, &length);
    int nargs = arglist.v.list[0].v.num;
    int fully = (nargs >= 2 && is_true(arglist.v.list[2]));
    Var r;
    int i;

    free_var(arglist);
    if (!bytes)
	return make_error_pack(E_INVARG);

    if (fully) {
	r = new_list(length);
	for (i = 1; i <= length; i++) {
	    r.v.list[i].type = TYPE_INT;
	    r.v.list[i].v.num = (unsigned char) bytes[i - 1];
	}
    } else {
	static Stream *s = 0;
	int count, in_string;

	if (!s)
	    s = new_stream(50);

	for (count = in_string = 0, i = 0; i < length; i++) {
	    unsigned char c = bytes[i];

	    if (isgraph(c) || c == ' ' || c == '\t') {
		if (!in_string)
		    count++;
		in_string = 1;
	    } else {
		count++;
		in_string = 0;
	    }
	}

	r = new_list(count);
	for (count = 1, in_string = 0, i = 0; i < length; i++) {
	    unsigned char c = bytes[i];

	    if (isgraph(c) || c == ' ' || c == '\t') {
		stream_add_char(s, c);
		in_string = 1;
	    } else {
		if (in_string) {
		    r.v.list[count].type = TYPE_STR;
		    r.v.list[count].v.str = str_dup(reset_stream(s));
		    count++;
		}
		r.v.list[count].type = TYPE_INT;
		r.v.list[count].v.num = c;
		count++;
		in_string = 0;
	    }
	}

	if (in_string) {
	    r.v.list[count].type = TYPE_STR;
	    r.v.list[count].v.str = str_dup(reset_stream(s));
	}
    }

    return make_var_pack(r);
}

static int
encode_binary(Stream * s, Var v)
{
    int i;

    switch (v.type) {
    case TYPE_INT:
	if (v.v.num < 0 || v.v.num >= 256)
	    return 0;
	stream_add_char(s, (char) v.v.num);
	break;
    case TYPE_STR:
	stream_add_string(s, v.v.str);
	break;
    case TYPE_LIST:
	for (i = 1; i <= v.v.list[0].v.num; i++)
	    if (!encode_binary(s, v.v.list[i]))
		return 0;
	break;
    default:
	return 0;
    }

    return 1;
}

static package
bf_encode_binary(Var arglist, Byte next, void *vdata, Objid progr)
{
    static Stream *s = 0;
    int ok, length;
    Var r;
    const char *bytes;

    if (!s)
	s = new_stream(100);

    ok = encode_binary(s, arglist);
    free_var(arglist);
    length = stream_length(s);
    bytes = reset_stream(s);
    if (ok) {
	r.type = TYPE_STR;
	r.v.str = str_dup(raw_bytes_to_binary(bytes, length));
	return make_var_pack(r);
    } else
	return make_error_pack(E_INVARG);
}

static package
bf_explode(Var arglist, Byte next, void *vdata, Objid progr)
{
	const char *delim = str_dup(" ");
	
	if (!is_wizard(progr))
		return make_raise_pack(E_PERM, "Temporarily requires wizperms (Security)", zero);
	
	if (arglist.v.list[0].v.num > 1)
		delim = arglist.v.list[2].v.str;
	
	return make_var_pack(strexplode(arglist.v.list[1].v.str, delim));
}

/* Character manipulation/conversion */

static package
bf_ucs_or_asc(Var arglist, Byte next, void *vdata, Objid progr, int isasc)
{	// bf_asc raises E_INVARG if over 255
	Var v;
	int complete = 0;

	if (arglist.v.list[0].v.num > 1) {
		if (!is_options_list(arglist.v.list[2])) {
			free_var(arglist);
			return make_error_pack(E_INVARG);
		}
		if(options_list_value(arglist.v.list[2], "complete", &v)) {
			complete = is_true(v);
			free_var(v);
		}
	}
	if (complete) {
		int offset = 0, len = strlen(arglist.v.list[1].v.str);
		Var tmp;
		
		v = new_list(0);
		tmp.type = TYPE_INT;
		while (offset < len) {
			tmp.v.num = utf8_to_ucs4((char *) (arglist.v.list[1].v.str + offset));
			if (tmp.v.num < 0 || (tmp.v.num > 255 && isasc)) {
				free_var(arglist);
				free_var(v);
				return make_error_pack(E_INVARG);
			}
			offset += utf8_numbytes(arglist.v.list[1].v.str[offset]);
			v = listappend(v, tmp);
		}
		free_var(arglist);
	} else {
		if (0 == arglist.v.list[1].v.str[0]) {
			/* empty string */
			free_var(arglist);
			return make_raise_pack(E_INVARG, "Empty String", zero);
		}
		v.type = TYPE_INT;
		v.v.num = utf8_to_ucs4((char *) arglist.v.list[1].v.str);
		free_var(arglist);
		if (v.v.num < 0 || (v.v.num > 255 && isasc))
			return make_error_pack(E_INVARG);
	}
	return make_var_pack(v);
}

static package
bf_ucs(Var arglist, Byte next, void *vdata, Objid progr)
{ return bf_ucs_or_asc(arglist, next, vdata, progr, 0); }

static package
bf_asc(Var arglist, Byte next, void *vdata, Objid progr)
{ return bf_ucs_or_asc(arglist, next, vdata, progr, 1); }

static package
bf_octet_int(Var arglist, Byte next, void *vdata, Objid progr)
{
	Var v;
	int complete = 0;

	if (arglist.v.list[0].v.num > 1) {
		if (!is_options_list(arglist.v.list[2])) {
			free_var(arglist);
			return make_error_pack(E_INVARG);
		}
		if(options_list_value(arglist.v.list[2], "complete", &v)) {
			complete = is_true(v);
			free_var(v);
		}
	}
	if (complete) {
		int len = strlen(arglist.v.list[1].v.str), i;
		
		v = new_list(len);
		for (i = 0; i < len; ++i) {
			v.v.list[i + 1].type = TYPE_INT;
			v.v.list[i + 1].v.num = (unsigned char) arglist.v.list[1].v.str[i];
		}
	} else {
		if (0 == arglist.v.list[1].v.str[0]) {
			/* empty string */
			free_var(arglist);
			return make_raise_pack(E_INVARG, "Empty String", zero);
		}
		v.type = TYPE_INT;
		v.v.num = (unsigned char) arglist.v.list[1].v.str[0];
	}
    free_var(arglist);
    return make_var_pack(v);
}

static package
bf_chr(Var arglist, Byte next, void *vdata, Objid progr)
{
	/* Note: This currently does not allow character 0 to be used */
	Var v;

	v.type = TYPE_STR;
	switch (arglist.v.list[1].type) {
	case TYPE_INT:
		if (1 > arglist.v.list[1].v.num) {
			free_var(arglist);
			return make_error_pack(E_RANGE);
		}
		v.v.str = ucs4_to_utf8(arglist.v.list[1].v.num);
		break;
	case TYPE_LIST: {
		int len = arglist.v.list[1].v.list[0].v.num, i;
		Stream *out = new_stream(len + 1);
		char *thischar;
		
		for (i = 1; i <= arglist.v.list[1].v.list[0].v.num; ++i) {
			if (arglist.v.list[1].v.list[i].type != TYPE_INT) {
				free_var(arglist);	free_stream(out);
				return make_error_pack(E_TYPE);
			} else if (1 > arglist.v.list[1].v.list[i].v.num) {
				free_var(arglist);	free_stream(out);
				return make_error_pack(E_RANGE);
			}
			thischar = ucs4_to_utf8(arglist.v.list[1].v.list[i].v.num);
			stream_add_string(out, thischar);
			free_str(thischar);
		}
		v.v.str = str_dup(reset_stream(out));
		break; }
	default:
		return make_error_pack(E_TYPE);
	}
	free_var(arglist);
	return make_var_pack(v);
}

static package
bf_octet_str(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var v;

    v.type = TYPE_STR;
	switch (arglist.v.list[1].type) {
	case TYPE_INT:
		if (arglist.v.list[1].v.num < 1 ||
			arglist.v.list[1].v.num > 0xFF) {
		/* not an integer in valid ASCII range */
			free_var(arglist);
			return make_error_pack(E_INVARG);
		}
		v.v.str = (char *) mymalloc(2, M_STRING);
		((char *) v.v.str)[0] = arglist.v.list[1].v.num;
		((char *) v.v.str)[1] = 0;
		break;
	case TYPE_LIST: {
		int len = arglist.v.list[1].v.list[0].v.num, i;
		
		v.v.str = (char *) mymalloc(len + 1, M_STRING);
		((char *) v.v.str)[len] = 0;
		for (i = 1; i <= arglist.v.list[1].v.list[0].v.num; ++i) {
			if (arglist.v.list[1].v.list[i].type != TYPE_INT) {
				free_var(arglist);	free_var(v);
				return make_error_pack(E_TYPE);
			} else if (arglist.v.list[1].v.list[i].v.num < 1 ||
					   arglist.v.list[1].v.list[i].v.num > 0xFF) {
				free_var(arglist);	free_var(v);
				return make_error_pack(E_RANGE);
			}
			((char *) v.v.str)[i - 1] = arglist.v.list[1].v.list[i].v.num;
		}
		break; }
	default:
		return make_error_pack(E_TYPE);
	}
    free_var(arglist);
    return make_var_pack(v);
}

void
register_list(void)
{
    register_function("value_bytes", 1, 1, bf_value_bytes, TYPE_ANY);
    register_function("value_hash", 1, 1, bf_value_hash, TYPE_ANY);
    register_function("string_hash", 1, 1, bf_string_hash, TYPE_STR);
    register_function("binary_hash", 1, 1, bf_binary_hash, TYPE_STR);
    register_function("decode_binary", 1, 2, bf_decode_binary,
		      TYPE_STR, TYPE_ANY);
    register_function("encode_binary", 0, -1, bf_encode_binary);
    register_function("octets", 1, 1, bf_octets, TYPE_STR);
    register_function("getoctets", 2, 3, bf_getoctets, TYPE_STR, TYPE_INT, TYPE_INT);
    register_function("getoctet", 2, 2, bf_getoctet, TYPE_STR, TYPE_INT);
    register_function("setoctets", 3, 4, bf_setoctets, TYPE_STR, TYPE_STR, TYPE_INT, TYPE_INT);
    register_function("setoctet", 3, 3, bf_setoctet, TYPE_STR, TYPE_STR, TYPE_INT);
    /* list */
    register_function("length", 1, 1, bf_length, TYPE_ANY);
    register_function("slice", 1, 2, bf_slice, TYPE_LIST, TYPE_ANY);
    register_function("remove_duplicates", 1, 1, bf_remove_duplicates, TYPE_LIST);
    register_function("setadd", 2, 2, bf_setadd, TYPE_LIST, TYPE_ANY);
    register_function("setremove", 2, 2, bf_setremove, TYPE_LIST, TYPE_ANY);
    register_function("enlist", 1, 1, bf_enlist, TYPE_ANY);
	register_function("make_list", 1, 2, bf_make_list, TYPE_INT, TYPE_ANY);
	register_function("make", 1, 2, bf_make_list, TYPE_INT, TYPE_ANY);	// For Martian-compliance
    register_function("sort", 1, 2, bf_sort, TYPE_LIST, TYPE_LIST);
    register_function("listappend", 2, 3, bf_listappend,
		      TYPE_LIST, TYPE_ANY, TYPE_INT);
    register_function("listinsert", 2, 3, bf_listinsert,
		      TYPE_LIST, TYPE_ANY, TYPE_INT);
    register_function("listdelete", 2, 2, bf_listdelete, TYPE_LIST, TYPE_INT);
    register_function("listset", 3, 3, bf_listset,
		      TYPE_LIST, TYPE_ANY, TYPE_INT);
    register_function("equal", 2, 2, bf_equal, TYPE_ANY, TYPE_ANY);
    register_function("is_member", 2, 2, bf_is_member, TYPE_ANY, TYPE_LIST);
    register_function("iassoc", 2, 3, bf_iassoc, TYPE_ANY, TYPE_LIST, TYPE_INT);
    register_function("assoc", 2, 3, bf_assoc, TYPE_ANY, TYPE_LIST, TYPE_INT);
    register_function("random_of", 1, 1, bf_random_of, TYPE_LIST);

    /* string */
    register_function("tostr", 0, -1, bf_tostr);
    register_function("toliteral", 1, 1, bf_toliteral, TYPE_ANY);
    setup_pattern_cache();
    register_function("match", 2, 3, bf_match, TYPE_STR, TYPE_STR, TYPE_ANY);
    register_function("rmatch", 2, 3, bf_rmatch, TYPE_STR, TYPE_STR, TYPE_ANY);
    register_function("substitute", 2, 2, bf_substitute, TYPE_STR, TYPE_LIST);
    register_function("crypt", 1, 2, bf_crypt, TYPE_STR, TYPE_STR);
    register_function("index", 2, 3, bf_index, TYPE_STR, TYPE_STR, TYPE_ANY);
    register_function("rindex", 2, 3, bf_rindex, TYPE_STR, TYPE_STR, TYPE_ANY);
    register_function("strcmp", 2, 2, bf_strcmp, TYPE_STR, TYPE_STR);
    register_function("strsub", 3, 4, bf_strsub,
		      TYPE_STR, TYPE_STR, TYPE_STR, TYPE_ANY);
	register_function("explode", 1, 2, bf_explode, TYPE_STR, TYPE_STR);
			  
	/* character */
	register_function("ucs4", 1, 2, bf_ucs, TYPE_STR, TYPE_LIST);
	register_function("ucs", 1, 2, bf_ucs, TYPE_STR, TYPE_LIST);
	register_function("asc", 1, 2, bf_asc, TYPE_STR, TYPE_LIST);
	register_function("octet_int", 1, 2, bf_octet_int, TYPE_STR);
	register_function("chr", 1, 1, bf_chr, TYPE_ANY);
	register_function("octet_str", 1, 1, bf_octet_str, TYPE_ANY);
}


char rcsid_list[] = "$Id: list.c,v 1.13 2004/06/22 01:25:07 luke-jr Exp $";

/* 
 * $Log: list.c,v $
 * Revision 1.13  2004/06/22 01:25:07  luke-jr
 * The `ucs4' built-in function has been renamed to `ucs'.
 * The `chr' and `octet_str' built-in functions will now accept a list of valid INTs and output a string. The `ucs', `asc', and `octet_int' built-in functions will output a LIST of INTs with the full content of the strings provided if the second argument is an option list with the option `complete' set to true.
 *
 * Revision 1.12  2004/03/27 03:42:01  luke-jr
 * Code cleanup neccesary to compile in Debian 2.2 using GCC 2.95.2 and Linux 2.4.19
 *
 * Revision 1.11  2004/03/27 01:38:52  luke-jr
 * Move string quoting to utils.c
 * Fix memory leak in pull_input (net_multi)
 *
 * Revision 1.10  2004/03/06 22:50:54  luke-jr
 * Unprintable characters, such as control and non-ASCII characters, can be placed within MOO strings using escape codes.
 * Common escape codes have been taken from C.
 * Control characters can be represented as \^A (for Ctrl-A, 0x01).
 * Specific unicode characters can be referenced by using \2xx, \4xxxx, or \8xxxxxxxx (where x is a hexadecimal digit).
 *
 * Revision 1.9  2004/02/25 02:52:06  luke-jr
 * sort() bifunc ported from JHCore; nothing left in listutils.c, deleted
 *
 * Revision 1.8  2004/02/14 06:42:46  luke-jr
 * Ported explode() from JHCore. listutils.c nearly replaced.
 *
 * Revision 1.7  2004/02/11 02:16:09  luke-jr
 * For security, certain advanced builtin functions (whose equivalent MOO code would run out of ticks) require wizard permissions: slice, remove_duplicates, make_list, iassoc, assoc, sort
 *
 * Revision 1.6  2004/02/11 00:46:32  luke-jr
 * MOO strings are now treated as UTF-8
 *
 * Revision 1.5  2004/02/04 02:25:42  luke-jr
 * Reported slice() from JHCore, reducing listutils to 2 (buggy) functions
 *
 * Revision 1.4  2004/02/03 01:03:08  luke-jr
 * Moved most of listutils to list
 *
 * Revision 1.3  2004/01/31 06:36:56  luke-jr
 * list_assoc moved to list.c
 *
 * Revision 1.2  2004/01/24 19:02:52  luke-jr
 * Moved substr (list.c)'s primary code to substr_real (utils.c) so we can use it with char *
 *
 * Revision 1.1.1.1  2004/01/18 20:43:12  luke-jr
 * Initial import
 *
 * Revision 1.6  2001/03/12 00:16:29  bjj
 * bf_crypt now passes the entire second argument as the salt to
 * the C crypt() routine.  This works fine for traditional DES crypts
 * and supports modern modular crypts like FreeBSD's.  This just makes
 * it possible to pass the entire salt:  the core still has to do it.
 *
 * Revision 1.5  1998/12/14 13:17:57  nop
 * Merge UNSAFE_OPTS (ref fixups); fix Log tag placement to fit CVS whims
 *
 * Revision 1.4  1997/07/07 03:24:54  nop
 * Merge UNSAFE_OPTS (r5) after extensive testing.
 * 
 * Revision 1.3.2.3  1997/07/03 08:04:01  bjj
 * Pattern cache was not storing case_matters flag, causing many patterns to
 * be impossible to find in the cache.
 *
 * Revision 1.3.2.2  1997/05/20 14:55:52  nop
 * Include Jason's patch for bf_decode_binary losing on systems where
 * char is signed.
 *
 * Revision 1.3.2.1  1997/03/21 15:22:56  bjj
 * doinsert reallocs for appending to refcnt 1 lists.  note that this wins
 * because it avoids all the var_ref/free_var that's done in the general case,
 * not because it avoids malloc/free.  the general case could also benefit from
 * using memcpy when the refcnt is 1, rather than looping with var_ref.
 *
 * Revision 1.3  1997/03/03 06:20:04  bjj
 * new_list(0) now returns the same empty list to every caller
 *
 * Revision 1.2  1997/03/03 04:18:46  nop
 * GNU Indent normalization
 *
 * Revision 1.1.1.1  1997/03/03 03:45:00  nop
 * LambdaMOO 1.8.0p5
 *
 * Revision 2.7  1996/03/11  23:35:17  pavel
 * Fixed bad use of possibly-signed characters in decode_binary().
 * Release 1.8.0p1.
 *
 * Revision 2.6  1996/02/18  23:17:24  pavel
 * Added value_hash(), string_hash(), and binary_hash().  Release 1.8.0beta3.
 *
 * Revision 2.5  1996/02/08  07:02:09  pavel
 * Added support for floating-point numbers.  Fixed registration of
 * decode_binary().  Renamed err/logf() to errlog/oklog() and TYPE_NUM to
 * TYPE_INT.  Updated copyright notice for 1996.  Release 1.8.0beta1.
 *
 * Revision 2.4  1996/01/16  07:26:56  pavel
 * Fixed `case_matters' arguments to strsub(), index(), rindex(), match(), and
 * rmatch() to allow values of any type.  Release 1.8.0alpha6.
 *
 * Revision 2.3  1996/01/11  07:42:14  pavel
 * Added support for C's crypt() function being unavailable.  Fixed potential
 * string overread in case of a too-short salt argument to MOO's crypt()
 * function.  Added built-ins encode_binary() and decode_binary(), in support
 * of new binary I/O facilities.  Release 1.8.0alpha5.
 *
 * Revision 2.2  1995/12/31  03:25:04  pavel
 * Added missing #include "options.h".  Release 1.8.0alpha4.
 *
 * Revision 2.1  1995/12/11  07:43:20  pavel
 * Moved value_bytes built-in function to here.
 *
 * Release 1.8.0alpha2.
 *
 * Revision 2.0  1995/11/30  04:24:03  pavel
 * New baseline version, corresponding to release 1.8.0alpha1.
 *
 * Revision 1.12  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.11  1992/10/23  19:27:07  pavel
 * Removed a place where a local structure variable was initialized in its
 * declaration, since some compilers can't hack that.
 * Added the `%%' -> `%' transformation to substitute().
 *
 * Revision 1.10  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.9  1992/10/17  20:33:17  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 * Added some (int) casts to placate over-protective compilers.
 *
 * Revision 1.8  1992/09/08  22:15:09  pjames
 * Updated strrangeset() and listrangeset() to use correct algorithm.
 *
 * Revision 1.7  1992/09/08  22:03:45  pjames
 * Added all code from bf_str_list.c, and some code from bf_type.c.
 *
 * Revision 1.6  1992/08/31  22:31:47  pjames
 * Changed some `char *'s to `const char *' and fixed code accordingly.
 *
 * Revision 1.5  1992/08/28  23:20:22  pjames
 * Added `listrangeset()' and `strrangeset()'.
 *
 * Revision 1.4  1992/08/28  16:29:10  pjames
 * Added some varref()'s.
 * Removed some free_var()'s, due to ref-counting.
 * Changed some my_free()'s to free_var().
 * Made `substr()' preserve the string it is passed.
 *
 * Revision 1.3  1992/08/10  16:53:46  pjames
 * Updated #includes.
 *
 * Revision 1.2  1992/07/21  00:04:04  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
